package main

import "fmt"

func MemoizationJoinOrdering (QG QueryGraph, JTC JoinTreeCreator) *Tree {
  n := uint(len(QG.R))
  BestTree := make([]*Tree, 1<<n) // 1<<n == 2^n
  for i := uint(0); i < n; i++ {
    BestTree[1<<i] = &Tree{float64(QG.R[i]), 1<<i, nil, nil, 0, nil}
  }
  allRel := uint((1<<n) - 1)
  return MemoizationJoinOrderingSub(allRel, &BestTree, QG, JTC)
}


func MemoizationJoinOrderingSub(S uint, BestTree *[]*Tree, QG QueryGraph, JTC JoinTreeCreator) *Tree {
  if (*BestTree)[S] != nil {
    return (*BestTree)[S]
  }
  for _, S1 := range Subsets(S) {
    n := uint(len(QG.R))
    S2 := SetMinus(S, S1, n)
    CurrTree := JTC.CreateJoinTree(MemoizationJoinOrderingSub(S1, BestTree, QG, JTC),
                                   MemoizationJoinOrderingSub(S2, BestTree, QG, JTC),
                                   QG)
    if (*BestTree)[S] == nil || (*BestTree)[S].Cost > CurrTree.Cost {
      (*BestTree)[S] = CurrTree
    }
  }
  return (*BestTree)[S]
}

func main () {
  var QG QueryGraph
  Costfunctions := []costfunction_t{Cnlj, Chj, Csmj}
  JTC := JoinTreeCreator{false, false, Costfunctions}
  QG.R = []uint{10, 20, 20, 10}
  QG.S = map[uint]float64{}
  QG.SetSel(0, 1, 0.01)
  QG.SetSel(1, 2, 0.5)
  QG.SetSel(2, 3, 0.01)

  resMemo := MemoizationJoinOrdering(QG, JTC)
  fmt.Println("MemoizationJoinOrdering:", resMemo.ToString())
}
